import sqlite3

import pandas as pd

from .utils import _extract_first_field


class BASEdb(object):
    def __init__(self, sqlite_file):
        """Initialize SRAdb.

        Parameters
        ----------

        sqlite_file: string
                     Path to unzipped SRAmetadb.sqlite file


        """
        self.sqlite_file = sqlite_file
        self.open()
        self.cursor = self.db.cursor()

    def open(self):
        """Open sqlite connection."""
        self.db = sqlite3.connect(self.sqlite_file)
        self.db.text_factory = str

    def close(self):
        """Close sqlite connection."""
        self.db.close()

    def list_tables(self):
        """List all tables in the sqlite file.


        Returns
        -------
        table_list: list
                    List of all table names
        """
        results = self.cursor.execute(
            'SELECT name FROM sqlite_master WHERE type="table";').fetchall()
        return _extract_first_field(results)

    def list_fields(self, table):
        """List all fields in a given table.

        Parameters
        ----------
        table: string
               Table name.
               See `list_tables` for getting all table names

        Returns
        -------
        field_list: list
                    A list of field names for the table
        """
        results = self.cursor.execute('SELECT * FROM {}'.format(table))
        return _extract_first_field(results.description)

    def desc_table(self, table):
        """Describe all fields in a table.

        Parameters
        ----------
        table: string
               Table name.
               See `list_tables` for getting all table names

        Returns
        -------
        table_desc: DataFrame
                    A DataFrame with field name and its
                    schema description
        """
        results = self.cursor.execute(
            'PRAGMA table_info("{}")'.format(table)).fetchall()
        columns = ['cid', 'name', 'dtype', 'notnull', 'dflt_value', 'pk']
        data = []
        for result in results:
            data.append(list(map(lambda x: str(x), result)))
        table_desc = pd.DataFrame(data, columns=columns)
        return table_desc

    def query(self, sql_query):
        """Run SQL query.

        Parameters
        ----------
        sql_query: string
                   SQL query string

        Returns
        -------
        results: DataFrame
                 Query results formatted as dataframe

        """
        results = self.cursor.execute(sql_query).fetchall()
        column_names = list(map(lambda x: x[0], self.cursor.description))
        results = [dict(zip(column_names, result)) for result in results]
        return pd.DataFrame(results)

    def get_row_count(self, table):
        """Get row counts for a table.

        Parameters
        ----------
        table: string
               Table name.
               See `list_tables` for getting all table names

        Returns
        -------
        row_count: int
                   Number of rows in table
        """
        return self.cursor.execute(
            'SELECT max(rowid) FROM {}'.format(table)).fetchone()[0]

    def all_row_counts(self):
        """Get row counts of all tables in the db file.

        Returns
        -------
        row_counts: DataFrame
                    A dataframe with table names and corresponding
                    row count.

        """
        tables = self.list_tables()
        results = dict(
            [(table, self.get_row_count(table)) for table in tables])
        return pd.DataFrame.from_dict(
            results, orient='index', columns=['count'])



"""Command line interface for pysradb
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os

from . import __version__
from .sradb import download_sradb_file
from .sradb import SRAdb
from .geodb import download_geodb_file
from .geodb import GEOdb

import click
import pandas as pd

click.disable_unicode_literals_warning = True
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


def _check_sradb_file(db):
    if db is None:
        db = os.path.join(os.getcwd(), 'SRAmetadb.sqlite')
        download_sradb_file()
    if not os.path.isfile(db):
        raise RuntimeError('{} does not exist'.format(db))
    return db


def _check_geodb_file(db):
    if db is None:
        db = os.path.join(os.getcwd(), 'GEOmetadb.sqlite')
        download_geodb_file()
    if not os.path.isfile(db):
        raise RuntimeError('{} does not exist'.format(db))
    return db


@click.group()
@click.version_option(version=__version__)
def cli():
    """pysradb: Query NGS metadata and data from NCBI Sequence Read Archive.

    Citation: Pending.
    """
    pass


@cli.command(
    'srametadb',
    context_settings=CONTEXT_SETTINGS,
    help='Download SRAmetadb.sqlite')
@click.option('--out_dir', type=str, help='Output directory location')
@click.option('--overwrite', type=bool, help='Overwrite existing file')
def cmd_download_sra(out_dir, overwrite):
    if out_dir is None:
        out_dir = os.getcwd()
    download_sradb_file(out_dir, overwrite)


@cli.command(
    'geometadb',
    context_settings=CONTEXT_SETTINGS,
    help='Download GEOmetadb.sqlite')
@click.option('--out_dir', type=str, help='Output directory location')
@click.option('--overwrite', type=bool, help='Overwrite existing file')
def cmd_download_geo(out_dir, overwrite):
    if out_dir is None:
        out_dir = os.getcwd()
    download_geodb_file(out_dir, overwrite)


@cli.command(
    'download',
    context_settings=CONTEXT_SETTINGS,
    help='Download SRA project (SRPnnnn)')
@click.option('--out_dir', help='Output directory root')
@click.option('--db', help='Path to SRAmetadb.sqlite file')
@click.argument('srp_ids', nargs=-1, required=True)
def cmd_download_sra(out_dir, db, srp_ids):
    db = _check_sradb_file(db)
    if out_dir is None:
        out_dir = os.path.join(os.getcwd(), 'pysradb_downloads')
    sradb = SRAdb(db)
    for srp in srp_ids:
        sradb.download(srp=srp, out_dir=out_dir)
    sradb.close()


@cli.command(
    'sra-metadata',
    context_settings=CONTEXT_SETTINGS,
    help='Fetch metadata for SRA project (SRPnnnn)')
@click.option('--saveto', help='Save metadata dataframe to file')
@click.option('--db', help='Path to SRAmetadb.sqlite file')
@click.option(
    '--expand',
    is_flag=True,
    help='Should sample_attribute be expanded',
    default=False)
@click.argument('srp_id', required=True)
def cmd_sra_metadata(srp_id, db, expand, saveto):
    db = _check_sradb_file(db)
    sradb = SRAdb(db)
    df = sradb.sra_metadata(acc=srp_id, expand_sample_attributes=expand)
    if saveto:
        df.to_csv(saveto, index=False, header=True, sep='\t')
    else:
        with pd.option_context('display.max_rows', None, 'display.max_columns',
                               None):
            if len(df.index):
                print(df)
    sradb.close()


@cli.command(
    'gse-metadata',
    context_settings=CONTEXT_SETTINGS,
    help='Fetch metadata for GEO ID (GSEnnnn)')
@click.option('--saveto', help='Save metadata dataframe to file')
@click.option('--db', help='Path to GEOmetadb.sqlite file')
@click.argument('gse_id', required=True)
def cmd_gse_metadata(gse_id, db, saveto):
    db = _check_geodb_file(db)
    geodb = GEOdb(db)
    df = geodb.gse_metadata(gse=gse_id)
    if saveto:
        df.to_csv(saveto, index=False, header=True, sep='\t')
    else:
        with pd.option_context('display.max_rows', None, 'display.max_columns',
                               None):
            if len(df.index):
                print(df)
    geodb.close()


@cli.command(
    'gsm-metadata',
    context_settings=CONTEXT_SETTINGS,
    help='Fetch metadata for GSM ID (GSMnnnn)')
@click.option('--saveto', help='Save metadata dataframe to file')
@click.option('--db', help='Path to GEOmetadb.sqlite file')
@click.argument('gsm_id', required=True)
def cmd_gsm_metadata(gsm_id, db, saveto):
    db = _check_geodb_file(db)
    geodb = GEOdb(db)
    df = geodb.gsm_metadata(gsm=gsm_id)
    if saveto:
        df.to_csv(saveto, index=False, header=True, sep='\t')
    else:
        with pd.option_context('display.max_rows', None, 'display.max_columns',
                               None):
            if len(df.index):
                print(df)
    geodb.close()


@cli.command(
    'gse-to-gsm', context_settings=CONTEXT_SETTINGS, help='Get GSM(s) for GSE')
@click.option('--saveto', help='Save metadata dataframe to file')
@click.option('--db', help='Path to GEOmetadb.sqlite file')
@click.argument('gse_id', required=True)
def cmd_gse_to_gsm(gse_id, db, saveto):
    db = _check_geodb_file(db)
    geodb = GEOdb(db)
    df = geodb.gse_to_gsm(gse=gse_id)
    if saveto:
        df.to_csv(saveto, index=False, header=True, sep='\t')
    else:
        with pd.option_context('display.max_rows', None, 'display.max_columns',
                               None):
            if len(df.index):
                print(df)
    geodb.close()



"""Tests for cli.py
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import os
from pysradb import cli as sradbcli
from pysradb import GEOdb, SRAdb

import pytest
from click.testing import CliRunner


@pytest.fixture(scope="module")
def sradb_connection(conf_download_sradb_file):
    db_file = conf_download_sradb_file
    db = SRAdb(db_file)
    return db


@pytest.fixture(scope="module")
def geodb_connection(conf_download_geodb_file):
    db_file = conf_download_geodb_file
    db = GEOdb(db_file)
    return db


@pytest.fixture(scope="module")
def runner():
    return CliRunner()


def test_all_row_counts_geo(geodb_connection):
    assert geodb_connection.all_row_counts().loc['metaInfo', 'count'] == 2


def test_all_row_counts_sra(sradb_connection):
    assert sradb_connection.all_row_counts().loc['metaInfo', 'count'] == 2


def test_download(runner):
    result = runner.invoke(sradbcli.cmd_download_sra, [
        '--db', 'data/SRAmetadb.sqlite', '--out_dir', 'srp_downloads',
        'SRP063852'
    ])
    assert os.path.getsize('srp_downloads/SRP063852/SRX1254413/SRR2433794.sra')


def test_sra_metadata(runner):
    result = runner.invoke(sradbcli.cmd_sra_metadata,
                           ['SRP098789', '--db', 'data/SRAmetadb.sqlite'])
    assert 'SRX2536403' in result.output

def test_sra_metadata(runner):
    result = runner.invoke(sradbcli.cmd_sra_metadata,
                           ['SRP098789', '--db', 'data/SRAmetadb.sqlite', '--expand'])
    assert 'treatment_time' in result.output


def test_gse_metadata(runner):
    result = runner.invoke(sradbcli.cmd_gse_metadata,
                           ['GSE114314', '--db', 'data/GEOmetadb.sqlite'])
    assert 'Name: Robert Ivanek' in result.output


def test_gsm_metadata(runner):
    result = runner.invoke(sradbcli.cmd_gsm_metadata,
                           ['GSM3139409', '--db', 'data/GEOmetadb.sqlite'])
    assert result.exit_code == 0


def test_gse_to_gsm(runner):
    result = runner.invoke(sradbcli.cmd_gse_to_gsm,
                           ['GSE114314', '--db', 'data/GEOmetadb.sqlite'])
    assert 'GSM3139411' in result.output


def test_cmd_gse_to_gsm_empty(runner):
    result = runner.invoke(sradbcli.cmd_gse_to_gsm, input='\n')
    assert 'Missing argument' in result.output
    assert result.exit_code == 2



"""Methods to interact with SRA"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import gzip
import os
import re
import sys
import warnings

import pandas as pd
from tqdm import tqdm

from .basedb import BASEdb

from .filter_attrs import expand_sample_attribute_columns

from .utils import _find_aspera_keypath
from .utils import _get_url
from .utils import get_gzip_uncompressed_size
from .utils import copyfileobj
from .utils import mkdir_p
from .utils import order_dataframe
from .utils import run_command

PY3 = True
if sys.version_info[0] < 3:
    PY3 = False

FTP_PREFIX = {
    'fasp': 'anonftp@ftp-trace.ncbi.nlm.nih.gov:',
    'ftp': 'ftp://ftp-trace.ncbi.nlm.nih.gov'
}
SRADB_URL = [
    'https://s3.amazonaws.com/starbuck1/sradb/SRAmetadb.sqlite.gz',
    'https://gbnci-abcc.ncifcrf.gov/backup/SRAmetadb.sqlite.gz'
]

ASCP_CMD_PREFIX = 'ascp -k 1 -QT -l 2000m -i'


def download_sradb_file(download_dir=os.getcwd(), overwrite=True):
    """Download SRAdb.sqlite file.

    Parameters
    ----------
    download_dir: string
                  Directory to download SRAmetadb.sqlite
    overwrite: bool
               overwrite existing file(s).
               Set to True by default.

    """
    download_location = os.path.join(download_dir, 'SRAmetadb.sqlite.gz')
    download_location_unzip = download_location.rstrip('.gz')

    if os.path.isfile(download_location) and overwrite is False:
        raise RuntimeError(
            '{} already exists! Set `overwrite=True` to redownload.'.forma(
                download_location))
    if os.path.isfile(download_location_unzip) and overwrite is False:
        raise RuntimeError(
            '{} already exists! Set `overwrite=True` to redownload.'.format(
                download_location_unzip))
    if os.path.isfile(download_location_unzip):
        os.remove(download_location_unzip)
    if os.path.isfile(download_location):
        os.remove(download_location)
    try:
        _get_url(SRADB_URL[0], download_location)
    except Exception as e:
        # Try other URL
        warnings.warn(
            'Could not use {}.\nException: {}.\nTrying alternate url ...\n'.
            format(SRADB_URL[0], e), RuntimeWarning)
        _get_url(SRADB_URL[1], download_location)
    print('Extracting {} ...'.format(download_location))
    filesize = get_gzip_uncompressed_size(download_location)
    with gzip.open(download_location, 'rb') as fh_in:
        with open(download_location_unzip, 'wb') as fh_out:
            copyfileobj(
                fh_in,
                fh_out,
                filesize=filesize,
                desc='Extracting {}'.format('SRAmetadb.sqlite.gz'))
    print('Done!')
    db = SRAdb(download_location_unzip)
    metadata = db.query('SELECT * FROM metaInfo')
    db.close()
    print('Metadata associated with {}:'.format(download_location_unzip))
    print(metadata)


class SRAdb(BASEdb):
    def __init__(self, sqlite_file):
        """Initialize SRAdb.

        Parameters
        ----------

        sqlite_file: string
                     Path to unzipped SRAmetadb.sqlite file


        """
        super(SRAdb, self).__init__(sqlite_file)
        self._db_type = 'SRA'
        self.valid_in_acc_type = [
            'SRA', 'ERA', 'DRA', 'SRP', 'ERP', 'DRP', 'SRS', 'ERS', 'DRS',
            'SRX', 'ERX', 'DRX', 'SRR', 'ERR', 'DRR'
        ]
        self.valid_in_type = {
            'SRA': 'submission',
            'ERA': 'submission',
            'DRA': 'submission',
            'SRP': 'study',
            'ERP': 'study',
            'DRP': 'study',
            'SRS': 'sample',
            'ERS': 'sample',
            'DRS': 'sample',
            'SRX': 'experiment',
            'ERX': 'experiment',
            'DRX': 'experiment',
            'SRR': 'run',
            'ERR': 'run',
            'DRR': 'run'
        }

    def sra_metadata(self,
                     acc,
                     out_type=[
                         'study_accession',
                         'experiment_accession',
                         'experiment_title',
                         'experiment_attribute',
                         'sample_attribute',
                         'run_accession',
                         'taxon_id',
                         'library_selection',
                         'library_layout',
                         'library_strategy',
                         'library_source',
                         'library_name',
                         'bases',
                         'spots',
                         'adapter_spec',
                     ],
                     expand_sample_attributes=False):
        """Get metadata for the provided SRA accession.

        Parameters
        ----------
        acc: string
             SRA accession ID
        out_type: list
                  List of columns to output
        expand_sample_attributes: bool
                                  Should sample_attribute column be expanded?

        Returns
        -------
        metadata_df: DataFrame
                     A dataframe with all relevant columns
        """
        in_acc_type = re.sub('\\d+$', '', acc).upper()
        if in_acc_type not in self.valid_in_acc_type:
            raise ValueError('{} not a valid input type'.format(in_acc_type))

        in_type = self.valid_in_type[in_acc_type]
        out_type = [x for x in out_type if x != in_type]

        select_type = [in_type + '_accession'] + out_type
        select_type_sql = (',').join(select_type)
        sql = "SELECT DISTINCT " + select_type_sql + \
              " FROM sra_ft WHERE sra_ft MATCH '" + acc + "';"
        df = self.query(sql)
        df['avg_read_length'] = df['bases'] / df['spots']
        df['spots'] = df['spots'].astype(int)
        df['bases'] = df['bases'].astype(int)
        df['taxon_id'] = df['taxon_id'].fillna(0).astype(int)
        df = df.sort_values(by=[
            'taxon_id', 'avg_read_length', 'run_accession',
            'experiment_accession', 'library_selection'
        ])
        metadata_df = df[out_type + ['avg_read_length']].reset_index(drop=True)
        if expand_sample_attributes:
            if 'sample_attribute' in metadata_df.columns:
                metadata_df = expand_sample_attribute_columns(metadata_df)
        return metadata_df

    def search_sra(
            self,
            search_str,
            out_type=[
                'study_accession', 'experiment_accession', 'experiment_title',
                'experiment_attribute', 'sample_attribute', 'run_accession',
                'taxon_id', 'library_selection', 'library_layout',
                'library_strategy', 'library_source', 'library_name', 'bases',
                'spots'
            ]):
        """Search SRA for any search term.

        Parameters
        ----------
        search_str: string
                    SQL like text string to search.
                    SQL like text => For example,
                    terms in quotes "" enforce an exact search.

        Returns
        -------
        query_df: DataFrame
                  Dataframe with relevant query results
        """

        select_type = out_type
        select_type_sql = (',').join(select_type)
        sql = "SELECT DISTINCT " + select_type_sql +\
              " FROM sra_ft WHERE sra_ft MATCH '" + search_str + "';"
        df = self.query(sql)
        if len(df.index) > 0:
            return order_dataframe(df, out_type)
        return None

    def search_by_expt_id(self, srx):
        """Search for a SRX/GSM id in the experiments.

        Parameters
        ----------
        srx: string
             SRX (experiment_accession) ID

        Returns
        -------
        results: dict
                 Dictionary with relevant hits
        """
        if 'GSM' in srx:
            results = self.cursor.execute(
                'select * from EXPERIMENT where experiment_alias = "{}"'.
                format(srx)).fetchall()
        else:
            results = self.cursor.execute(
                'select * from EXPERIMENT where experiment_accession = "{}"'.
                format(srx)).fetchall()
        assert len(results) == 1, 'Got multiple hits'
        results = results[0]
        column_names = list(map(lambda x: x[0], self.cursor.description))
        results = dict(zip(column_names, results))
        return pd.DataFrame.from_dict(results, orient='index').T

    def download(self,
                 srp=None,
                 df=None,
                 out_dir=None,
                 protocol='fasp',
                 ascp_dir=None):
        """Download SRA files.

        Parameters
        ----------
        srp: string
             SRP ID (optional)
        df: Dataframe
            A dataframe as obtained from `sra_metadata`
        out_dir: string
                 Directory location for download
        protocol: string
                  ['fasp'/'ftp'] fasp => faster download, ftp => slower
        ascp_dir: string
                  Location of ascp directory
        """
        if out_dir is None:
            out_dir = os.path.join(os.getcwd(), 'pysradb_downloads')
        if srp:
            df = self.sra_metadata(srp)
        if protocol == 'ftp':
            warnings.warn(
                '''Using `ftp` protocol leads to slower downloads.\n
                Consider using `fasp` after installing aspera-client.''',
                UserWarning)
        if protocol == 'fasp':
            if ascp_dir is None:
                ascp_dir = os.path.join(os.path.expanduser('~'), '.aspera')
            if not os.path.exists(ascp_dir):
                raise RuntimeError('''Count not find aspera at: {}\n
                    Install aspera-client following instructions
                    in the README.rst OR set `protocol`=ftp.\n'''.format(
                    ascp_dir))
            ascp_bin = os.path.join(ascp_dir, 'connect', 'bin', 'ascp')
        df = df.copy()
        df.loc[:, 'download_url'] = FTP_PREFIX[
            protocol] + '/sra/sra-instant/reads/ByRun/sra/' + df[
                'run_accession'].str[:3] + '/' + df[
                    'run_accession'].str[:6] + '/' + df[
                        'run_accession'] + '/' + df['run_accession'] + '.sra'
        download_list = df[[
            'study_accession', 'experiment_accession', 'run_accession',
            'download_url'
        ]].values
        with tqdm(total=download_list.shape[0]) as pbar:
            for srp, srx, srr, url in download_list:
                pbar.set_description('{}/{}/{}'.format(srp, srx, srr))
                srp_dir = os.path.join(out_dir, srp)
                srx_dir = os.path.join(srp_dir, srx)
                srr_location = os.path.join(srx_dir, srr + '.sra')
                mkdir_p(srx_dir)
                if protocol == 'fasp':
                    cmd = ASCP_CMD_PREFIX.replace('ascp', ascp_bin)
                    cmd = '{} {} {} {}'.format(cmd,
                                               _find_aspera_keypath(ascp_dir),
                                               url, srx_dir)
                    run_command(cmd, verbose=False)
                else:
                    _get_url(url, srr_location, show_progress=False)
                pbar.update()
        return df



"""Tests for geodb.py
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import pytest
from pysradb import GEOdb


@pytest.fixture(scope="module")
def geodb_connection(conf_download_geodb_file):
    db_file = conf_download_geodb_file
    db = GEOdb(db_file)
    return db


def test_all_row_counts(geodb_connection):
    assert geodb_connection.all_row_counts().loc['metaInfo', 'count'] == 2


def test_gse_metadata(geodb_connection):
    df = geodb_connection.gse_metadata('GSE114314')
    assert int(df['pubmed_id'][0]) == 29925996


def test_gse_to_gsm(geodb_connection):
    df = geodb_connection.gse_to_gsm('GSE114314')
    assert df['gsm'][3] == 'GSM3139412'


def test_geo_convert(geodb_connection):
    df = geodb_connection.geo_convert('GSM3139409')
    assert df['to_acc'][0] == 'GSE114314'


def test_guess_srp_form_gse(geodb_connection):
    srp = geodb_connection.guess_srp_from_gse('GSE73136')
    assert srp == 'SRP063852'



"""Tests for sradb.py
"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import os
import pytest
from pysradb import SRAdb
from pysradb.filter_attrs import guess_cell_type, guess_tissue_type, guess_strain_type


@pytest.fixture(scope="module")
def sradb_connection(conf_download_sradb_file):
    db_file = conf_download_sradb_file
    db = SRAdb(db_file)
    return db


def test_list_tables(sradb_connection):
    sra_tables = sradb_connection.list_tables()
    assert sra_tables == [
        'metaInfo', 'submission', 'study', 'sample', 'experiment', 'run',
        'sra', 'sra_ft', 'sra_ft_content', 'sra_ft_segments', 'sra_ft_segdir',
        'col_desc', 'fastq'
    ]


def test_list_fields(sradb_connection):
    fields = sradb_connection.list_fields('study')
    assert fields == [
        'study_ID', 'study_alias', 'study_accession', 'study_title',
        'study_type', 'study_abstract', 'broker_name', 'center_name',
        'center_project_name', 'study_description', 'related_studies',
        'primary_study', 'sra_link', 'study_url_link', 'xref_link',
        'study_entrez_link', 'ddbj_link', 'ena_link', 'study_attribute',
        'submission_accession', 'sradb_updated'
    ]


def test_desc_table(sradb_connection):
    names = sorted(sradb_connection.desc_table('sra_ft').name.tolist())
    assert names[:7] == [
        'SRR_bamFile', 'SRX_bamFile', 'SRX_fastqFTP', 'adapter_spec',
        'anonymized_name', 'base_caller', 'bases'
    ]


def test_all_row_counts(sradb_connection):
    assert sradb_connection.all_row_counts().loc['metaInfo', 'count'] == 2


def test_sra_metadata(sradb_connection):
    df = sradb_connection.sra_metadata('SRP017942')
    assert df['experiment_accession'][0] == 'SRX217028'


def test_sra_metadata2(sradb_connection):
    df = sradb_connection.sra_metadata(
        'SRP017942', expand_sample_attributes=True)
    assert df['transfected_with'][0] == '3xflag-gfp'


def test_search(sradb_connection):
    df = sradb_connection.search_sra(search_str="breast cancer")
    assert len(df.index)


def test_search_by_expt_id(sradb_connection):
    df = sradb_connection.search_by_expt_id('SRX1254413')
    assert df.study_name.tolist()[0] == 'GSE73136'


def test_download_fasp(sradb_connection):
    df = sradb_connection.sra_metadata('SRP098789')
    df = df[df.experiment_accession == 'SRX2536403']
    sradb_connection.download(df=df, out_dir='data/')
    assert os.path.isfile('data/SRP098789/SRX2536403/SRR5227288.sra')
    assert os.path.getsize('data/SRP098789/SRX2536403/SRR5227288.sra')
    os.remove('data/SRP098789/SRX2536403/SRR5227288.sra')

@pytest.mark.xfail
def test_download_ftp(sradb_connection):
    # This happens to fail because of ftp problems
    df = sradb_connection.sra_metadata('SRP098789')
    df = df[df.experiment_accession == 'SRX2536404']
    sradb_connection.download(df=df, protocol='ftp', out_dir='data/')
    assert os.path.isfile('data/SRP098789/SRX2536404/SRR5227289.sra')
    assert os.path.getsize('data/SRP098789/SRX2536404/SRR5227289.sra')
    os.remove('data/SRP098789/SRX2536404/SRR5227289.sra')

def test_tissue_type(sradb_connection):
    df = sradb_connection.sra_metadata('SRP016501')
    df = df[df.experiment_accession == 'SRX196389']
    cell_type = df['sample_attribute'].apply(lambda x: guess_cell_type(x))
    tissue_type = df['sample_attribute'].apply(lambda x: guess_tissue_type(x))
    assert cell_type.tolist() == ['chicken_brain']
    assert tissue_type.tolist() == ['brain']


def test_strain_type(sradb_connection):
    df = sradb_connection.sra_metadata('SRP043036')
    df = df.sort_values(by='experiment_accession')
    strains = df['sample_attribute'].apply(
        lambda x: guess_strain_type(x)).tolist()
    assert strains == [
        'by4741', 'by4741', 'by4741', 'by4741', 'by4741', 'by4741', 'by4741',
        'by4741', 's288c', 's288c', 's288c', 's288c'
    ]



# -*- coding: utf-8 -*-
"""Top-level package for pysradb."""

__author__ = """Saket Choudhary"""
__email__ = 'saketkc@gmail.com'
__version__ = '0.4.0'

from .sradb import download_sradb_file
from .sradb import SRAdb
from .geodb import download_geodb_file
from .geodb import GEOdb
from .filter_attrs import expand_sample_attribute_columns



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""The setup script."""

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['pandas>=0.23.4', 'tqdm>=4.28', 'click>=7.0']
test_requirements = [
    'pytest',
]

setup(
    author='Saket Choudhary',
    author_email='saketkc@gmail.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description=
    "Python package for interacting with SRAdb and downloading datasets from SRA",
    entry_points={'console_scripts': ['pysradb=pysradb.cli:cli']},
    install_requires=requirements,
    license="BSD license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='pysradb',
    name='pysradb',
    packages=['pysradb'],
    setup_requires=requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/saketkc/pysradb',
    version='0.4.0',
    zip_safe=False,
)



"""Methods to interact with SRA"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import gzip
import os
import re
import sys

from .utils import _get_url
from .utils import get_gzip_uncompressed_size
from .utils import copyfileobj
from .basedb import BASEdb

PY3 = True
if sys.version_info[0] < 3:
    PY3 = False

GEOmetadb_URL = 'http://starbuck1.s3.amazonaws.com/sradb/GEOmetadb.sqlite.gz'


def download_geodb_file(download_dir=os.getcwd(), overwrite=True):
    """Download GEOmetadb.sqlite file.

    Parameters
    ----------
    download_dir: string
                  Directory to download SRAmetadb.sqlite
    overwrite: bool
               overwrite existing file(s).
               Set to True by default.

    """
    download_location = os.path.join(download_dir, 'GEOmetadb.sqlite.gz')
    download_location_unzip = download_location.rstrip('.gz')

    if os.path.isfile(download_location) and overwrite is False:
        raise RuntimeError(
            '{} already exists! Set `overwrite=True` to redownload.'.forma(
                download_location))
    if os.path.isfile(download_location_unzip) and overwrite is False:
        raise RuntimeError(
            '{} already exists! Set `overwrite=True` to redownload.'.format(
                download_location_unzip))

    try:
        _get_url(GEOmetadb_URL, download_location)
    except Exception as e:
        raise RuntimeError('Could not use {}.\nException: {}.\n'.format(
            GEOmetadb_URL, e))
    print('Extracting {} ...'.format(download_location))
    filesize = get_gzip_uncompressed_size(download_location)
    with gzip.open(download_location, 'rb') as fh_in:
        with open(download_location_unzip, 'wb') as fh_out:
            copyfileobj(
                fh_in,
                fh_out,
                filesize=filesize,
                desc='Extracting {}'.format('GEOmetadb.sqlite.gz'))
    print('Done!')
    db = GEOdb(download_location_unzip)
    metadata = db.query('SELECT * FROM metaInfo')
    db.close()
    print('Metadata associated with {}:'.format(download_location_unzip))
    print(metadata)


class GEOdb(BASEdb):
    def __init__(self, sqlite_file):
        """Initialize SRAdb.

        Parameters
        ----------

        sqlite_file: string
                     Path to unzipped SRAmetadb.sqlite file


        """
        super(GEOdb, self).__init__(sqlite_file)
        self._db_type = 'GEO'
        self.valid_in_type = ['GSE', 'GPL', 'GSM', 'GDS']

    def gse_metadata(self, gse):
        """Get metadata for GSE ID.

        Parameters
        ----------
        gse: string
             GSE ID

        Returns
        -------
        metadata_df: DataFrame
                     A dataframe with relevant fields
        """
        return self.query("SELECT * from gse WHERE gse='{}';".format(gse))

    def gsm_metadata(self, gsm):
        """Get metadata for GSM ID.

        Parameters
        ----------
        gsm: string
             GSM ID

        Returns
        -------
        metadata_df: DataFrame
                     A dataframe with relevant fields
        """
        return self.query("SELECT * from gsm WHERE gsm='{}';".format(gsm))

    def geo_convert(self, from_acc):
        """Convert one GEO accession to other.

        Parameters
        ----------
        from_acc: string
                  GPL/GSE/GSM accession ID

        Returns
        -------
        mapping_df: DataFrame
                    A dataframe with relevant mappings
        """
        return self.query(
            "SELECT * FROM geoConvert WHERE from_acc='{}';".format(from_acc))

    def gse_to_gsm(self, gse):
        """Fetch GSMs for a GSE.

        Parameters
        ----------
        gse: string
             GSE ID

        Returns
        -------
        mapping_df: DataFrame
                    A dataframe with relevant mappings
        """
        return self.query("SELECT * FROM gse_gsm WHERE gse='{}'".format(gse))

    def gsm_to_gse(self, gsm):
        """Fetch GSE for a GSM.

        Parameters
        ----------
        gsm: string
             GSM ID

        Returns
        -------
        mapping_df: DataFrame
                    A dataframe with relevant mappings
        """
        mapping_df = self.query(
            "SELECT * FROM gse_gsm WHERE gsm='{}'".format(gsm))
        return mapping_df.loc[:, ['gsm', 'gse']]

    def guess_srp_from_gse(self, gse):
        """Convert GSE to SRP id.

        Parameters
        ----------
        gse: string
             GSE ID

        Returns
        -------
        srp: string
             SRP ID
        """
        results = self.query('SELECT * FROM gse WHERE gse = "' + gse + '"')
        if results.shape[0] == 1:
            supp_file = results['supplementary_file'][0]
            if supp_file:
                splitted = supp_file.split(';')
                if len(splitted):
                    match = re.findall('SRP.*', splitted[-1])
                    if len(match):
                        srp = match[0].split('/')[-1]
                        return srp
        return None
